@@ -2,8 +2,7 @@ import inspect
 import json as json_module
 import logging
 from collections import namedtuple
-from collections.abc import Sequence
-from collections.abc import Sized
+from functools import partialmethod
 from functools import wraps
 from http import client
 from itertools import groupby
@@ -18,15 +17,19 @@ from typing import Iterator
 from typing import List
 from typing import Mapping
 from typing import Optional
+from typing import Sequence
+from typing import Sized
 from typing import Tuple
 from typing import Type
 from typing import Union
 from typing import overload
 from warnings import warn
 
+import toml as _toml
 from requests.adapters import HTTPAdapter
 from requests.adapters import MaxRetryError
 from requests.exceptions import ConnectionError
+from requests.exceptions import RetryError
 
 from responses.matchers import json_params_matcher as _json_params_matcher
 from responses.matchers import query_string_matcher as _query_string_matcher
@@ -63,7 +66,9 @@ from urllib.parse import urlunsplit
 
 if TYPE_CHECKING:  # pragma: no cover
     # import only for linter run
-    from mypy.typeshed.stdlib.unittest.mock import _patcher as _mock_patcher
+    import os
+    from unittest.mock import _patch as _mock_patcher
+
     from requests import PreparedRequest
     from requests import models
     from urllib3 import Retry as _Retry
@@ -181,9 +186,6 @@ def get_wrapped(
         Wrapped function
 
     """
-    if registry is not None:
-        responses._set_registry(registry)
-
     assert_mock = std_mock.patch.object(
         target=responses,
         attribute="assert_all_requests_are_fired",
@@ -195,6 +197,9 @@ def get_wrapped(
         @wraps(func)
         async def wrapper(*args: Any, **kwargs: Any) -> Any:  # type: ignore[misc]
 
+            if registry is not None:
+                responses._set_registry(registry)
+
             with assert_mock, responses:
                 return await func(*args, **kwargs)
 
@@ -203,6 +208,9 @@ def get_wrapped(
         @wraps(func)
         def wrapper(*args: Any, **kwargs: Any) -> Any:  # type: ignore[misc]
 
+            if registry is not None:
+                responses._set_registry(registry)
+
             with assert_mock, responses:
                 # set 'assert_all_requests_are_fired' temporarily for a single run.
                 # Mock automatically unsets to avoid leakage to another decorated
@@ -212,7 +220,7 @@ def get_wrapped(
     return wrapper
 
 
-class CallList(Sequence, Sized):  # type: ignore[type-arg]
+class CallList(Sequence[Any], Sized):
     def __init__(self) -> None:
         self._calls: List[Call] = []
 
@@ -222,7 +230,7 @@ class CallList(Sequence, Sized):  # type: ignore[type-arg]
     def __len__(self) -> int:
         return len(self._calls)
 
-    def __getitem__(self, idx: int) -> Call:  # type: ignore[override]
+    def __getitem__(self, idx: Union[int, slice]) -> Union[Call, List[Call]]:
         return self._calls[idx]
 
     def add(self, request: "PreparedRequest", response: _Body) -> None:
@@ -354,6 +362,8 @@ class BaseResponse(object):
         url: _URLPatternType,
         match_querystring: Union[bool, object] = None,
         match: "_MatcherIterable" = (),
+        *,
+        passthrough: bool = False,
     ) -> None:
         self.method: str = method
         # ensure the url has a default path set if the url is a string
@@ -366,6 +376,7 @@ class BaseResponse(object):
 
         self.match: "_MatcherIterable" = match
         self.call_count: int = 0
+        self.passthrough = passthrough
 
     def __eq__(self, other: Any) -> bool:
         if not isinstance(other, BaseResponse):
@@ -407,6 +418,25 @@ class BaseResponse(object):
         return bool(urlsplit(self.url).query)
 
     def _url_matches(self, url: _URLPatternType, other: str) -> bool:
+        """Compares two URLs.
+
+        Compares only scheme, netloc and path. If 'url' is a re.Pattern, then checks that
+        'other' matches the pattern.
+
+        Parameters
+        ----------
+        url : Union["Pattern[str]", str]
+            Reference URL or Pattern to compare.
+
+        other : str
+            URl that should be compared.
+
+        Returns
+        -------
+        bool
+            True, if URLs are identical or 'other' matches the pattern.
+
+        """
         if isinstance(url, str):
             if _has_unicode(url):
                 url = _clean_unicode(url)
@@ -455,6 +485,32 @@ class BaseResponse(object):
         return True, ""
 
 
+def _form_response(
+    body: Union[BufferedReader, BytesIO],
+    headers: Optional[Mapping[str, str]],
+    status: int,
+) -> HTTPResponse:
+    # The requests library's cookie handling depends on the response object
+    # having an original response object with the headers as the `msg`, so
+    # we give it what it needs.
+    data = BytesIO()
+    data.close()
+
+    orig_response = HTTPResponse(
+        body=data,  # required to avoid "ValueError: Unable to determine whether fp is closed."
+        msg=headers,
+        preload_content=False,
+    )
+    return HTTPResponse(
+        status=status,
+        reason=client.responses.get(status, None),
+        body=body,
+        headers=headers,
+        original_response=orig_response,
+        preload_content=False,
+    )
+
+
 class Response(BaseResponse):
     def __init__(
         self,
@@ -516,14 +572,7 @@ class Response(BaseResponse):
             content_length = len(body.getvalue())
             headers["Content-Length"] = str(content_length)
 
-        return HTTPResponse(
-            status=status,
-            reason=client.responses.get(status, None),
-            body=body,
-            headers=headers,
-            original_response=OriginalResponseShim(headers),
-            preload_content=False,
-        )
+        return _form_response(body, headers, status)
 
     def __repr__(self) -> str:
         return (
@@ -585,41 +634,12 @@ class CallbackResponse(BaseResponse):
         body = _handle_body(body)
         headers.extend(r_headers)
 
-        return HTTPResponse(
-            status=status,
-            reason=client.responses.get(status, None),
-            body=body,
-            headers=headers,
-            original_response=OriginalResponseShim(headers),
-            preload_content=False,
-        )
+        return _form_response(body, headers, status)
 
 
 class PassthroughResponse(BaseResponse):
-    passthrough: bool = True
-
-
-class OriginalResponseShim(object):
-    """
-    Shim for compatibility with older versions of urllib3
-
-    requests cookie handling depends on responses having a property chain of
-    `response._original_response.msg` which contains the response headers [1]
-
-    Using HTTPResponse() for this purpose causes compatibility errors with
-    urllib3<1.23.0. To avoid adding more dependencies we can use this shim.
-
-    [1]: https://github.com/psf/requests/blob/75bdc998e2d/requests/cookies.py#L125
-    """
-
-    def __init__(self, headers: Any) -> None:
-        self.msg: Any = headers
-
-    def isclosed(self) -> bool:
-        return True
-
-    def close(self) -> None:
-        return
+    def __init__(self, *args: Any, **kwargs: Any):
+        super().__init__(*args, passthrough=True, **kwargs)
 
 
 class RequestsMock(object):
@@ -648,13 +668,29 @@ class RequestsMock(object):
         self.response_callback: Optional[Callable[[Any], Response]] = response_callback
         self.passthru_prefixes: Tuple[_URLPatternType, ...] = tuple(passthru_prefixes)
         self.target: str = target
-        self._patcher: Optional["_mock_patcher"] = None
+        self._patcher: Optional["_mock_patcher[Any]"] = None
         self._thread_lock = _ThreadingLock()
 
     def get_registry(self) -> FirstMatchRegistry:
+        """Returns current registry instance with responses.
+
+        Returns
+        -------
+        FirstMatchRegistry
+            Current registry instance with responses.
+
+        """
         return self._registry
 
     def _set_registry(self, new_registry: Type[FirstMatchRegistry]) -> None:
+        """Replaces current registry with `new_registry`.
+
+        Parameters
+        ----------
+        new_registry : Type[FirstMatchRegistry]
+            Class reference of the registry that should be set, eg OrderedRegistry
+
+        """
         if self.registered():
             err_msg = (
                 "Cannot replace Registry, current registry has responses.\n"
@@ -665,6 +701,7 @@ class RequestsMock(object):
         self._registry = new_registry()
 
     def reset(self) -> None:
+        """Resets registry (including type), calls, passthru_prefixes to default values."""
         self._registry = FirstMatchRegistry()
         self._calls.reset()
         self.passthru_prefixes = ()
@@ -724,26 +761,28 @@ class RequestsMock(object):
         response = Response(method=method, url=url, body=body, **kwargs)
         return self._registry.add(response)
 
-    def delete(self, *args: Any, **kwargs: Any) -> BaseResponse:
-        return self.add(DELETE, *args, **kwargs)
+    delete = partialmethod(add, DELETE)
+    get = partialmethod(add, GET)
+    head = partialmethod(add, HEAD)
+    options = partialmethod(add, OPTIONS)
+    patch = partialmethod(add, PATCH)
+    post = partialmethod(add, POST)
+    put = partialmethod(add, PUT)
 
-    def get(self, *args: Any, **kwargs: Any) -> BaseResponse:
-        return self.add(GET, *args, **kwargs)
+    def _add_from_file(self, file_path: "Union[str, bytes, os.PathLike[Any]]") -> None:
+        with open(file_path) as file:
+            data = _toml.load(file)
 
-    def head(self, *args: Any, **kwargs: Any) -> BaseResponse:
-        return self.add(HEAD, *args, **kwargs)
-
-    def options(self, *args: Any, **kwargs: Any) -> BaseResponse:
-        return self.add(OPTIONS, *args, **kwargs)
-
-    def patch(self, *args: Any, **kwargs: Any) -> BaseResponse:
-        return self.add(PATCH, *args, **kwargs)
-
-    def post(self, *args: Any, **kwargs: Any) -> BaseResponse:
-        return self.add(POST, *args, **kwargs)
-
-    def put(self, *args: Any, **kwargs: Any) -> BaseResponse:
-        return self.add(PUT, *args, **kwargs)
+        for rsp in data["responses"]:
+            rsp = rsp["response"]
+            self.add(
+                method=rsp["method"],
+                url=rsp["url"],
+                body=rsp["body"],
+                status=rsp["status"],
+                content_type=rsp["content_type"],
+                auto_calculate_content_length=rsp["auto_calculate_content_length"],
+            )
 
     def add_passthru(self, prefix: _URLPatternType) -> None:
         """
@@ -1020,12 +1059,21 @@ class RequestsMock(object):
                 retries = retries.increment(
                     method=response.request.method,  # type: ignore[misc]
                     url=response.url,  # type: ignore[misc]
-                    response=response,  # type: ignore[misc]
+                    response=response.raw,  # type: ignore[misc]
                 )
                 return self._on_request(adapter, request, retries=retries, **kwargs)
-            except MaxRetryError:
+            except MaxRetryError as e:
                 if retries.raise_on_status:
-                    raise
+                    """Since we call 'retries.increment()' by ourselves, we always set "error"
+                    argument equal to None, thus, MaxRetryError exception will be raised with
+                    ResponseError as a 'reason'.
+
+                    Here we're emulating the `if isinstance(e.reason, ResponseError):`
+                    branch found at:
+                    https://github.com/psf/requests/blob/
+                    177dd90f18a8f4dc79a7d2049f0a3f4fcc5932a0/requests/adapters.py#L549"""
+                    raise RetryError(e, request=request)
+
                 return response
         return response
 
@@ -1092,6 +1140,7 @@ __all__ = [
     # Exposed by the RequestsMock class:
     "activate",
     "add",
+    "_add_from_file",
     "add_callback",
     "add_passthru",
     "_deprecated_assert_all_requests_are_fired",
@@ -1126,6 +1175,7 @@ __all__ = [
 # expose only methods and/or read-only methods
 activate = _default_mock.activate
 add = _default_mock.add
+_add_from_file = _default_mock._add_from_file
 add_callback = _default_mock.add_callback
 add_passthru = _default_mock.add_passthru
 _deprecated_assert_all_requests_are_fired = _default_mock.assert_all_requests_are_fired
